@@ -13,6 +13,7 @@
 #  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 #  See the License for the specific language governing permissions and
 #  limitations under the License.
+from collections import OrderedDict
 
 from itertools import chain
 
@@ -22,11 +23,11 @@ import time
 import stat
 from calculate.core.server.gen_pid import search_worked_process2
 from calculate.core.setup_package import ChainProgressTemplate
-from calculate.lib.cl_template import templateFunction
+from calculate.lib.cl_template import templateFunction, SystemIni
 from calculate.lib.datavars import DataVars, Variable
-from calculate.lib.utils.colortext import (TextState, get_color_print,
-                                           convert_console_to_xml)
+from calculate.lib.utils.colortext import (TextState, get_color_print)
 from calculate.lib.utils.common import CmdlineParams
+from calculate.lib.utils.portage import BinaryPackage, get_binary_file
 from calculate.lib.utils.files import (
     pathJoin, PercentProgress, getProgPath, process, STDOUT, removeDir,
     makeDirectory, writeFile, readLinesFile, chmod, chown, FilePermission,
@@ -35,18 +36,24 @@ from calculate.lib.utils.files import (
 from calculate.lib.utils.git import Git
 from calculate.lib.utils.portage import (Layman, EmergeLog, EmergeLogNamedTask,
                                          InstalledPackageInfo, EbuildInfoError,
-                                         EbuildInfo, ChrootEix,
+                                         EbuildInfo, ChrootEix, getRequires,
                                          get_packages_files_directory,
-                                         get_manifest_files_directory)
+                                         get_manifest_files_directory,
+                                         PackageList, VDB_PATH,
+                                         hide_packages, unhide_packages,
+                                         LibraryProviders,
+                                         PackageError, getInstalledAtom)
 from calculate.update.emerge_parser import (EmergeParser,
                                             EmergeError, EmergeCommand, Chroot,
                                             Linux32, CommandExecutor,
                                             RevdepPercentBlock, EmergeCache)
 from calculate.lib.cl_log import log
+from calculate.update.update_tasks import EmergeMark
 from .build_storage import Build
 from calculate.update.update import Update
 from calculate.install.distr import (Distributive, IsoDistributive,
                                      DistributiveError)
+import shutil
 from functools import partial
 import os
 from os import path
@@ -81,6 +88,8 @@ class Builder(Update):
         self.update_map = {}
         self.color_print = get_color_print()
         self.emerge_cache = EmergeCache()
+        self.pkgnum = None
+        self.pkgnummax = None
 
     def mount_target(self, target):
         dir_distro = target.convertToDirectory()
@@ -297,10 +306,12 @@ class Builder(Update):
         with self.clVars.useDefaultModule("update"):
             return super(Builder, self).regenCache(repname)
 
-    def syncRepositories(self, repname, clean_on_error=True):
+    def syncRepositories(self, repname, fallback_sync=False,
+                         clean_on_error=True):
         with self.clVars.useDefaultModule("update"):
             return super(Builder, self).syncRepositories(
-                repname, clean_on_error=clean_on_error)
+                repname, fallback_sync=fallback_sync,
+                clean_on_error=clean_on_error)
 
     def _regenCache_process(self, progname, repname, cpu_num):
         chroot_path = self.clVars.Get('builder.cl_builder_path')
@@ -379,7 +390,7 @@ class Builder(Update):
             self._startEmerging(emerge)
         return True
 
-    def depclean(self):
+    def depclean(self, without_bdeps=False):
         """
         Выполнить очистку системы от лишних пакетов
         """
@@ -387,7 +398,8 @@ class Builder(Update):
         chroot_path = self.clVars.Get('cl_builder_path')
         logfile = self._get_log_file()
         with EmergeParser(self.chrootize(chroot_path, EmergeCommand(
-                ["--depclean", "--ask=y"],
+                (["--depclean", "--with-bdeps=n", "--ask=y"]
+                 if without_bdeps else ["--depclean", "--ask=y"]),
                 logfile=logfile,
                 emerge_default_opts=deo))) as emerge:
             try:
@@ -710,13 +722,12 @@ class Builder(Update):
         :return:
         """
         eselect_command = "/usr/bin/eselect"
-        p = self.chroot_process(builder_path, eselect_command, "--colour=yes",
+        p = self.chroot_process(builder_path, eselect_command, "--colour=no",
                                 "news", "list", stderr=STDOUT)
-        self.printPre(convert_console_to_xml(
-            p.read()).replace("  ", "&nbsp;&nbsp;"))
-        p = self.chroot_process(builder_path, eselect_command, "--colour=yes",
-                                "news", "read", "new", stderr=STDOUT)
-        return p.success()
+        re_new = re.compile(r"^\s+\[\d+\]\s+N?\s+(\S+)\s+(.*)$", re.M)
+        for date, title in reversed(re_new.findall(p.read())[-3:]):
+            self.printSUCCESS("{date} {title}".format(date=date, title=title))
+        return True
 
     def check_obsolete(self, builder_path):
         """
@@ -1013,9 +1024,11 @@ class Builder(Update):
         deo = self.get_default_emerge_opts()
         extra_params = [x for x in params if x.startswith("-")]
         packages = [x for x in params if not x.startswith("-")]
+        logfile = self._get_log_file()
         with EmergeParser(self.chrootize(builder_path, EmergeCommand(
                 packages, emerge_default_opts=deo,
-                extra_params=extra_params, use=use))) as emerge:
+                extra_params=extra_params, use=use,
+                logfile=logfile))) as emerge:
             try:
                 emerge.question.action = lambda x: False
                 emerge.run()
@@ -1324,3 +1337,216 @@ class Builder(Update):
             self.endTask()
         return True
 
+    def pretend_emerge_list(self, builder_path, *packages):
+        """
+        Получить список устанавливаемых пакетов
+        :param builder_path:
+        :param packages:
+        :return:
+        """
+        deo = self.get_default_emerge_opts()
+        logfile = self._get_log_file()
+        with EmergeParser(self.chrootize(builder_path, EmergeCommand(
+                        ["=%s" % x for x in packages] + ["@system"],
+                emerge_default_opts=deo,
+                extra_params=["-pKve"], logfile=logfile))) as emerge:
+            try:
+                emerge.question.action = lambda x: False
+                emerge.run()
+                for pkg in emerge.install_packages.list:
+                    yield pkg
+            except EmergeError:
+                self._display_error(emerge.prepare_error)
+
+    def check_automagic(self, builder_path):
+        """
+        Проверка на наличие неописанных автоматических зависимостей
+        :param builder_path:
+        :return:
+        """
+        task = EmergeLog(EmergeLogNamedTask(EmergeMark.Automagic),
+                         prefix=builder_path)
+
+        cache_list = ("/var/calculate/tmp/%s.checkdep" %
+                      self.clVars.Get("cl_builder_id_path"))
+        task_list = list(chain(*[list(getInstalledAtom(x, prefix=builder_path))
+                                 for x in readLinesFile(cache_list,
+                                                        grab=True)]))
+        task_list = list(task.list) + task_list
+
+        vdb_path = "var/db/.pkg"
+        real_vdb_path = path.join(builder_path, VDB_PATH)
+        hide_vdb_path = path.join(builder_path, vdb_path)
+        try:
+            shutil.move(real_vdb_path, hide_vdb_path)
+        except (IOError, OSError):
+            raise BuilderError(_("Failed to hide package database"))
+
+        automagic = OrderedDict()
+        automagic_waste = {}
+        automagic_clear = {}
+        try:
+            system_ini = SystemIni(self.clVars.Get('cl_builder_linux_datavars'))
+            lp = LibraryProviders(vdb_path=vdb_path, prefix=builder_path)
+
+            def get_check_data():
+                for pkg in PackageList(task_list):
+                    def get_all_reqs(pkg):
+                        for arch, lib in getRequires(pkg, vdb_path=vdb_path,
+                                                     prefix=builder_path):
+                            if arch in lp and lib in lp[arch]:
+                                yield tuple(lp[arch][lib])
+
+                    yield pkg, list(set(get_all_reqs(pkg)))
+
+            check_data = {x: y for x, y in get_check_data()
+                          if x["CATEGORY"] != "virtual"}
+
+            for i, data in enumerate(sorted(check_data.items(),
+                                            key=lambda x:x[0])):
+                package, required_pkgs = data
+
+                self.startTask("Check (%d of %d) %s" % (
+                    i + 1, len(check_data), package))
+
+                if not required_pkgs:
+                    self.endTask(True)
+                    continue
+
+                pretend = list(self.pretend_emerge_list(builder_path, package))
+                if not pretend:
+                    self.endTask(False)
+                    self.printERROR("Failed to receive package list")
+                else:
+                    # получаем список пакетов которые необходимых для работы
+                    required_pkgs = [x for x in required_pkgs
+                                     if all(y not in pretend for y in x)]
+                    required_pkgs = list(set(chain(*required_pkgs)))
+
+                    clear_req_pkgs = filter(
+                        None, [x.strip() for x in system_ini.getVar(
+                            "automagic-clear",
+                            package["CATEGORY/PN"]).split(",")])
+                    waste_pkgs = [x for x in clear_req_pkgs if
+                                  all(y["CATEGORY/PN"] != x
+                                      for y in required_pkgs)]
+                    if waste_pkgs:
+                        automagic_waste[package] = waste_pkgs
+                    automagic[package] = []
+                    automagic_clear[package] = []
+                    for pkg in required_pkgs:
+                        if any(x == pkg["CATEGORY/PN"] for x in clear_req_pkgs):
+                            automagic_clear[package].append(pkg)
+                        else:
+                            automagic[package].append(pkg)
+                    self.endTask(True)
+        finally:
+            try:
+                shutil.move(hide_vdb_path, real_vdb_path)
+            except (IOError, OSError):
+                raise BuilderError(_("Failed to unhide package database"))
+        self._report_automagic(automagic, automagic_clear, automagic_waste)
+        self._rebuild_automagic(automagic, automagic_clear, builder_path)
+        return True
+
+    def _report_automagic(self, automagic, automagic_clear, automagic_waste):
+        """
+        Сообщить информацию о вычисленных автоматических зависимостей
+        :param automagic:
+        :param automagic_clear:
+        :param automagic_waste:
+        :return:
+        """
+        for pkg, reqs in automagic.items():
+            if reqs:
+                self.printWARNING(
+                    "Auto depends for %s package by %s will be used" % (
+                        pkg, ",".join(str(x) for x in reqs)))
+        for pkg, clear_req_pkgs in automagic_clear.items():
+            if clear_req_pkgs:
+                self.printWARNING(
+                    "Auto depends for %s package by %s will be clear" % (
+                        pkg, ",".join(str(x) for x in clear_req_pkgs)))
+        for pkg, waste_pkgs in automagic_waste.items():
+            if waste_pkgs:
+                self.printWARNING(
+                    "Specified depends for %s package by %s obsoleted" % (
+                        pkg, ",".join(waste_pkgs)))
+
+    def _rebuild_automagic(self, automagic, automagic_clear, builder_path):
+        """
+        Пересобрать пакеты с автоматическими зависимостями
+        :param automagic:
+        :param automagic_clear:
+        :param builder_path:
+        :return:
+        """
+        automagic_log_dn = "/var/log/calculate/automagic"
+        makeDirectory(automagic_log_dn)
+        self.pkgnum = 0
+        self.pkgnummax = sum(1 for k, v in automagic.items()
+                             if v or automagic_clear[k])
+        work_dn = '/var/calculate/tmp/xpak-%s' % self.clVars.Get(
+            'cl_builder_id_path')
+
+        try:
+            for pkg, reqs in automagic.items():
+                clear_req_pkgs = automagic_clear.get(pkg, [])
+                if not reqs and not clear_req_pkgs:
+                    continue
+
+                self.pkgnum += 1
+                if clear_req_pkgs:
+                    try:
+                        hide_packages(*clear_req_pkgs, prefix=builder_path)
+
+                        self.emerge(builder_path, "-O", "=%s" % pkg)
+                    except EmergeError:
+                        old_logfile = self._get_log_file()
+                        pkg_path = str(pkg).replace("/", "_")
+                        new_logfile = '%s/%s-%s.log' % (
+                            automagic_log_dn,
+                            self.clVars.Get('cl_builder_id_path'), pkg_path)
+                        try:
+                            os.rename(old_logfile, new_logfile)
+                        except (OSError, IOError) as e:
+                            self.printERROR("{message}: {error}".format(
+                                message=_("Failed to save build log"),
+                                error=str(e)))
+                        self.printERROR(
+                            _("Failed to merge {package} without "
+                              "{hidden_pkgs}").format(
+                                package=str(pkg),
+                                hidden_pkgs=",".join(str(x) for x in reqs)))
+                        raise
+                    except PackageError as e:
+                        raise BuilderError(str(e))
+                    finally:
+                        try:
+                            unhide_packages(prefix=builder_path, force=True)
+                        except PackageError as e:
+                            raise BuilderError(str(e))
+                if reqs:
+                    self.printWARNING(
+                        _("({num} of {nummax}) Inject dependences for "
+                          "{package} package").format(num=self.pkgnum,
+                                                       nummax=self.pkgnummax,
+                                                       package=str(pkg)))
+                    pkg_fn = get_binary_file(
+                        pkg, self.clVars.Get('cl_builder_linux_pkgdir'))
+                    bp = BinaryPackage(pkg_fn, work_dn)
+                    try:
+                        bp["RDEPEND"] = "{oldpkgs} {newpkgs}".format(
+                            oldpkgs=bp["RDEPEND"],
+                            newpkgs=" ".join("%s:%s" % (
+                                x["CATEGORY/PN"], x["SLOT"]) for x in reqs))
+                        bp["autodeps"] = "\n".join("%s:%s" % (
+                            x["CATEGORY/PN"], x["SLOT"]) for x in reqs)
+                        bp.save()
+                    finally:
+                        bp.clear()
+        finally:
+            self.pkgnummax = None
+            self.pkgnum = None
+
+        return True
